@@ -37,7 +37,7 @@
 #ifdef ENABLE_HEARTBEAT
 /**
   * gnutls_heartbeat_enable:
-  * @session: is a #gnutls_session_t structure.
+  * @session: is a #gnutls_session_t type.
   * @type: one of the GNUTLS_HB_* flags
   *
   * If this function is called with the %GNUTLS_HB_PEER_ALLOWED_TO_SEND
@@ -56,14 +56,14 @@ void gnutls_heartbeat_enable(gnutls_session_t session, unsigned int type)
 {
 	extension_priv_data_t epriv;
 
-	epriv.num = type;
+	epriv = (void*)(intptr_t)type;
 	_gnutls_ext_set_session_data(session, GNUTLS_EXTENSION_HEARTBEAT,
 				     epriv);
 }
 
 /**
   * gnutls_heartbeat_allowed:
-  * @session: is a #gnutls_session_t structure.
+  * @session: is a #gnutls_session_t type.
   * @type: one of %GNUTLS_HB_LOCAL_ALLOWED_TO_SEND and %GNUTLS_HB_PEER_ALLOWED_TO_SEND
   *
   * This function will check whether heartbeats are allowed
@@ -77,14 +77,17 @@ int gnutls_heartbeat_allowed(gnutls_session_t session, unsigned int type)
 {
 	extension_priv_data_t epriv;
 
+	if (session->internals.handshake_in_progress != 0)
+		return 0; /* not allowed */
+
 	if (_gnutls_ext_get_session_data
 	    (session, GNUTLS_EXTENSION_HEARTBEAT, &epriv) < 0)
 		return 0;	/* Not enabled */
 
 	if (type == GNUTLS_HB_LOCAL_ALLOWED_TO_SEND) {
-		if (epriv.num & LOCAL_ALLOWED_TO_SEND)
+		if (((intptr_t)epriv) & LOCAL_ALLOWED_TO_SEND)
 			return 1;
-	} else if (epriv.num & GNUTLS_HB_PEER_ALLOWED_TO_SEND)
+	} else if (((intptr_t)epriv) & GNUTLS_HB_PEER_ALLOWED_TO_SEND)
 		return 1;
 
 	return 0;
@@ -136,7 +139,7 @@ heartbeat_send_data(gnutls_session_t session, const void *data,
 
 /**
  * gnutls_heartbeat_ping:
- * @session: is a #gnutls_session_t structure.
+ * @session: is a #gnutls_session_t type.
  * @data_size: is the length of the ping payload.
  * @max_tries: if flags is %GNUTLS_HEARTBEAT_WAIT then this sets the number of retransmissions. Use zero for indefinite (until timeout).
  * @flags: if %GNUTLS_HEARTBEAT_WAIT then wait for pong or timeout instead of returning immediately.
@@ -227,7 +230,7 @@ gnutls_heartbeat_ping(gnutls_session_t session, size_t data_size,
 	case SHB_RECV:
 		ret =
 		    _gnutls_recv_int(session, GNUTLS_HEARTBEAT, -1, NULL,
-				     0, NULL,
+				     NULL, 0, NULL,
 				     session->internals.
 				     hb_actual_retrans_timeout_ms);
 		if (ret == GNUTLS_E_HEARTBEAT_PONG_RECEIVED) {
@@ -269,7 +272,7 @@ gnutls_heartbeat_ping(gnutls_session_t session, size_t data_size,
 
 /**
  * gnutls_heartbeat_pong:
- * @session: is a #gnutls_session_t structure.
+ * @session: is a #gnutls_session_t type.
  * @flags: should be zero
  *
  * This function replies to a ping by sending a pong to the peer.
@@ -382,7 +385,7 @@ int _gnutls_heartbeat_handle(gnutls_session_t session, mbuffer_st * bufel)
 
 /**
  * gnutls_heartbeat_get_timeout:
- * @session: is a #gnutls_session_t structure.
+ * @session: is a #gnutls_session_t type.
  *
  * This function will return the milliseconds remaining
  * for a retransmission of the previously sent ping
@@ -410,7 +413,7 @@ unsigned int gnutls_heartbeat_get_timeout(gnutls_session_t session)
 
 /**
  * gnutls_heartbeat_set_timeouts:
- * @session: is a #gnutls_session_t structure.
+ * @session: is a #gnutls_session_t type.
  * @retrans_timeout: The time at which a retransmission will occur in milliseconds
  * @total_timeout: The time at which the connection will be aborted, in milliseconds.
  *
@@ -453,7 +456,7 @@ _gnutls_heartbeat_recv_params(gnutls_session_t session,
 	if (_data_size == 0)
 		return GNUTLS_E_UNEXPECTED_PACKET_LENGTH;
 
-	policy = epriv.num;
+	policy = (intptr_t)epriv;
 
 	if (data[0] == 1)
 		policy |= LOCAL_ALLOWED_TO_SEND;
@@ -463,7 +466,7 @@ _gnutls_heartbeat_recv_params(gnutls_session_t session,
 		return
 		    gnutls_assert_val(GNUTLS_E_RECEIVED_ILLEGAL_PARAMETER);
 
-	epriv.num = policy;
+	epriv = (void*)(intptr_t)policy;
 	_gnutls_ext_set_session_data(session, GNUTLS_EXTENSION_HEARTBEAT,
 				     epriv);
 
@@ -481,7 +484,7 @@ _gnutls_heartbeat_send_params(gnutls_session_t session,
 	    (session, GNUTLS_EXTENSION_HEARTBEAT, &epriv) < 0)
 		return 0;	/* nothing to send - not enabled */
 
-	if (epriv.num & GNUTLS_HB_PEER_ALLOWED_TO_SEND)
+	if (((intptr_t)epriv) & GNUTLS_HB_PEER_ALLOWED_TO_SEND)
 		p = 1;
 	else			/*if (epriv.num & GNUTLS_HB_PEER_NOT_ALLOWED_TO_SEND) */
 		p = 2;
@@ -497,7 +500,7 @@ _gnutls_heartbeat_pack(extension_priv_data_t epriv, gnutls_buffer_st * ps)
 {
 	int ret;
 
-	BUFFER_APPEND_NUM(ps, epriv.num);
+	BUFFER_APPEND_NUM(ps, (intptr_t)epriv);
 
 	return 0;
 
@@ -510,7 +513,7 @@ _gnutls_heartbeat_unpack(gnutls_buffer_st * ps,
 	extension_priv_data_t epriv;
 	int ret;
 
-	BUFFER_POP_NUM(ps, epriv.num);
+	BUFFER_POP_CAST_NUM(ps, epriv);
 
 	*_priv = epriv;
 
