@@ -47,9 +47,9 @@ int pubkey_to_bits(gnutls_pk_algorithm_t pk, gnutls_pk_params_st * params)
 {
 	switch (pk) {
 	case GNUTLS_PK_RSA:
-		return _gnutls_mpi_get_nbits(params->params[0]);
+		return _gnutls_mpi_get_nbits(params->params[RSA_MODULUS]);
 	case GNUTLS_PK_DSA:
-		return _gnutls_mpi_get_nbits(params->params[3]);
+		return _gnutls_mpi_get_nbits(params->params[DSA_P]);
 	case GNUTLS_PK_EC:
 		return gnutls_ecc_curve_get_size(params->flags) * 8;
 	default:
@@ -285,6 +285,7 @@ gnutls_pubkey_get_preferred_hash_algorithm(gnutls_pubkey_t key,
 	case GNUTLS_PK_DSA:
 		if (mand)
 			*mand = 1;
+		/* fallthrough */
 	case GNUTLS_PK_EC:
 
 		me = _gnutls_dsa_q_to_hash(key->pk_algorithm, &key->params, NULL);
@@ -763,7 +764,7 @@ gnutls_pubkey_export2(gnutls_pubkey_t key,
 /**
  * gnutls_pubkey_get_key_id:
  * @key: Holds the public key
- * @flags: should be 0 for now
+ * @flags: should be one of the flags from %gnutls_keyid_flags_t
  * @output_data: will contain the key ID
  * @output_data_size: holds the size of output_data (and will be
  *   replaced by the actual size of parameters)
@@ -796,7 +797,7 @@ gnutls_pubkey_get_key_id(gnutls_pubkey_t key, unsigned int flags,
 
 	ret =
 	    _gnutls_get_key_id(key->pk_algorithm, &key->params,
-			       output_data, output_data_size);
+			       output_data, output_data_size, flags);
 	if (ret < 0) {
 		gnutls_assert();
 		return ret;
@@ -808,13 +809,15 @@ gnutls_pubkey_get_key_id(gnutls_pubkey_t key, unsigned int flags,
 /**
  * gnutls_pubkey_export_rsa_raw:
  * @key: Holds the certificate
- * @m: will hold the modulus
- * @e: will hold the public exponent
+ * @m: will hold the modulus (may be %NULL)
+ * @e: will hold the public exponent (may be %NULL)
  *
  * This function will export the RSA public key's parameters found in
  * the given structure.  The new parameters will be allocated using
  * gnutls_malloc() and will be stored in the appropriate datum.
  *
+ * This function allows for %NULL parameters since 3.4.1.
+ *
  * Returns: %GNUTLS_E_SUCCESS on success, otherwise a negative error code.
  *
  * Since: 3.3.0
@@ -835,17 +838,21 @@ gnutls_pubkey_export_rsa_raw(gnutls_pubkey_t key,
 		return GNUTLS_E_INVALID_REQUEST;
 	}
 
-	ret = _gnutls_mpi_dprint_lz(key->params.params[0], m);
-	if (ret < 0) {
-		gnutls_assert();
-		return ret;
+	if (m) {
+		ret = _gnutls_mpi_dprint_lz(key->params.params[0], m);
+		if (ret < 0) {
+			gnutls_assert();
+			return ret;
+		}
 	}
 
-	ret = _gnutls_mpi_dprint_lz(key->params.params[1], e);
-	if (ret < 0) {
-		gnutls_assert();
-		_gnutls_free_datum(m);
-		return ret;
+	if (e) {
+		ret = _gnutls_mpi_dprint_lz(key->params.params[1], e);
+		if (ret < 0) {
+			gnutls_assert();
+			_gnutls_free_datum(m);
+			return ret;
+		}
 	}
 
 	return 0;
@@ -855,15 +862,17 @@ gnutls_pubkey_export_rsa_raw(gnutls_pubkey_t key,
 /**
  * gnutls_pubkey_export_dsa_raw:
  * @key: Holds the public key
- * @p: will hold the p
- * @q: will hold the q
- * @g: will hold the g
- * @y: will hold the y
+ * @p: will hold the p (may be %NULL)
+ * @q: will hold the q (may be %NULL)
+ * @g: will hold the g (may be %NULL)
+ * @y: will hold the y (may be %NULL)
  *
  * This function will export the DSA public key's parameters found in
  * the given certificate.  The new parameters will be allocated using
  * gnutls_malloc() and will be stored in the appropriate datum.
  *
+ * This function allows for %NULL parameters since 3.4.1.
+ *
  * Returns: %GNUTLS_E_SUCCESS on success, otherwise a negative error code.
  *
  * Since: 3.3.0
@@ -886,39 +895,45 @@ gnutls_pubkey_export_dsa_raw(gnutls_pubkey_t key,
 	}
 
 	/* P */
-	ret = _gnutls_mpi_dprint_lz(key->params.params[0], p);
-	if (ret < 0) {
-		gnutls_assert();
-		return ret;
+	if (p) {
+		ret = _gnutls_mpi_dprint_lz(key->params.params[0], p);
+		if (ret < 0) {
+			gnutls_assert();
+			return ret;
+		}
 	}
 
 	/* Q */
-	ret = _gnutls_mpi_dprint_lz(key->params.params[1], q);
-	if (ret < 0) {
-		gnutls_assert();
-		_gnutls_free_datum(p);
-		return ret;
+	if (q) {
+		ret = _gnutls_mpi_dprint_lz(key->params.params[1], q);
+		if (ret < 0) {
+			gnutls_assert();
+			_gnutls_free_datum(p);
+			return ret;
+		}
 	}
 
-
 	/* G */
-	ret = _gnutls_mpi_dprint_lz(key->params.params[2], g);
-	if (ret < 0) {
-		gnutls_assert();
-		_gnutls_free_datum(p);
-		_gnutls_free_datum(q);
-		return ret;
+	if (g) {
+		ret = _gnutls_mpi_dprint_lz(key->params.params[2], g);
+		if (ret < 0) {
+			gnutls_assert();
+			_gnutls_free_datum(p);
+			_gnutls_free_datum(q);
+			return ret;
+		}
 	}
 
-
 	/* Y */
-	ret = _gnutls_mpi_dprint_lz(key->params.params[3], y);
-	if (ret < 0) {
-		gnutls_assert();
-		_gnutls_free_datum(p);
-		_gnutls_free_datum(g);
-		_gnutls_free_datum(q);
-		return ret;
+	if (y) {
+		ret = _gnutls_mpi_dprint_lz(key->params.params[3], y);
+		if (ret < 0) {
+			gnutls_assert();
+			_gnutls_free_datum(p);
+			_gnutls_free_datum(g);
+			_gnutls_free_datum(q);
+			return ret;
+		}
 	}
 
 	return 0;
@@ -927,14 +942,16 @@ gnutls_pubkey_export_dsa_raw(gnutls_pubkey_t key,
 /**
  * gnutls_pubkey_export_ecc_raw:
  * @key: Holds the public key
- * @curve: will hold the curve
- * @x: will hold x
- * @y: will hold y
+ * @curve: will hold the curve (may be %NULL)
+ * @x: will hold x (may be %NULL)
+ * @y: will hold y (may be %NULL)
  *
  * This function will export the ECC public key's parameters found in
- * the given certificate.  The new parameters will be allocated using
+ * the given key.  The new parameters will be allocated using
  * gnutls_malloc() and will be stored in the appropriate datum.
  *
+ * This function allows for %NULL parameters since 3.4.1.
+ *
  * Returns: %GNUTLS_E_SUCCESS on success, otherwise a negative error code.
  *
  * Since: 3.0
@@ -956,21 +973,26 @@ gnutls_pubkey_export_ecc_raw(gnutls_pubkey_t key,
 		return GNUTLS_E_INVALID_REQUEST;
 	}
 
-	*curve = key->params.flags;
+	if (curve)
+		*curve = key->params.flags;
 
 	/* X */
-	ret = _gnutls_mpi_dprint_lz(key->params.params[ECC_X], x);
-	if (ret < 0) {
-		gnutls_assert();
-		return ret;
+	if (x) {
+		ret = _gnutls_mpi_dprint_lz(key->params.params[ECC_X], x);
+		if (ret < 0) {
+			gnutls_assert();
+			return ret;
+		}
 	}
 
 	/* Y */
-	ret = _gnutls_mpi_dprint_lz(key->params.params[ECC_Y], y);
-	if (ret < 0) {
-		gnutls_assert();
-		_gnutls_free_datum(x);
-		return ret;
+	if (y) {
+		ret = _gnutls_mpi_dprint_lz(key->params.params[ECC_Y], y);
+		if (ret < 0) {
+			gnutls_assert();
+			_gnutls_free_datum(x);
+			return ret;
+		}
 	}
 
 	return 0;
@@ -1559,11 +1581,13 @@ gnutls_pubkey_import_dsa_raw(gnutls_pubkey_t key,
 
 }
 
+#define OLD_PUBKEY_VERIFY_FLAG_TLS1_RSA 1
+
 /**
  * gnutls_pubkey_verify_data2:
  * @pubkey: Holds the public key
  * @algo: The signature algorithm used
- * @flags: Zero or one of %gnutls_pubkey_flags_t
+ * @flags: Zero or an OR list of #gnutls_certificate_verify_flags
  * @data: holds the signed data
  * @signature: contains the signature
  *
@@ -1571,7 +1595,9 @@ gnutls_pubkey_import_dsa_raw(gnutls_pubkey_t key,
  * parameters from the certificate.
  *
  * Returns: In case of a verification failure %GNUTLS_E_PK_SIG_VERIFY_FAILED 
- * is returned, and zero or positive code on success.
+ * is returned, and zero or positive code on success. For known to be insecure
+ * signatures this function will return %GNUTLS_E_INSUFFICIENT_SECURITY unless
+ * the flag %GNUTLS_VERIFY_ALLOW_BROKEN is specified.
  *
  * Since: 3.0
  **/
@@ -1590,7 +1616,7 @@ gnutls_pubkey_verify_data2(gnutls_pubkey_t pubkey,
 		return GNUTLS_E_INVALID_REQUEST;
 	}
 
-	if (flags & GNUTLS_PUBKEY_VERIFY_FLAG_TLS1_RSA)
+	if (flags & OLD_PUBKEY_VERIFY_FLAG_TLS1_RSA || flags & GNUTLS_VERIFY_USE_TLS1_RSA)
 		return gnutls_assert_val(GNUTLS_E_INVALID_REQUEST);
 
 	me = hash_to_entry(gnutls_sign_get_hash_algorithm(algo));
@@ -1601,8 +1627,17 @@ gnutls_pubkey_verify_data2(gnutls_pubkey_t pubkey,
 				 data, signature, &pubkey->params);
 	if (ret < 0) {
 		gnutls_assert();
+		return ret;
 	}
 
+	if (!(flags & GNUTLS_VERIFY_ALLOW_BROKEN)) {
+		if (gnutls_sign_is_secure(algo) == 0) {
+			return gnutls_assert_val(GNUTLS_E_INSUFFICIENT_SECURITY);
+		}
+	}
+
+	if (ret >= 0)
+		return 0;
 	return ret;
 }
 
@@ -1610,7 +1645,7 @@ gnutls_pubkey_verify_data2(gnutls_pubkey_t pubkey,
  * gnutls_pubkey_verify_hash2:
  * @key: Holds the public key
  * @algo: The signature algorithm used
- * @flags: Zero or one of %gnutls_pubkey_flags_t
+ * @flags: Zero or an OR list of #gnutls_certificate_verify_flags
  * @hash: holds the hash digest to be verified
  * @signature: contains the signature
  *
@@ -1638,7 +1673,7 @@ gnutls_pubkey_verify_hash2(gnutls_pubkey_t key,
 		return GNUTLS_E_INVALID_REQUEST;
 	}
 
-	if (flags & GNUTLS_PUBKEY_VERIFY_FLAG_TLS1_RSA) {
+	if (flags & OLD_PUBKEY_VERIFY_FLAG_TLS1_RSA || flags & GNUTLS_VERIFY_USE_TLS1_RSA) {
 		return _gnutls_pk_verify(GNUTLS_PK_RSA, hash, signature,
 					 &key->params);
 	} else {
