@@ -298,7 +298,7 @@ _gnutls_privkey_decode_ecc_key(ASN1_TYPE* pkey_asn, const gnutls_datum_t * raw_k
 			goto error;
 		}
 
-		pkey->params.flags = _gnutls_oid_to_ecc_curve(oid);
+		pkey->params.flags = gnutls_oid_to_ecc_curve(oid);
 
 		if (pkey->params.flags == GNUTLS_ECC_CURVE_INVALID) {
 			_gnutls_debug_log("Curve %s is not supported\n", oid);
@@ -470,33 +470,55 @@ gnutls_x509_privkey_import(gnutls_x509_privkey_t key,
 	/* If the Certificate is in PEM format then decode it
 	 */
 	if (format == GNUTLS_X509_FMT_PEM) {
-		/* Try the first header */
-		result =
-		    _gnutls_fbase64_decode(PEM_KEY_RSA, data->data,
-					   data->size, &_data);
+		unsigned left;
+		char *ptr;
+		uint8_t *begin_ptr;
 
-		if (result >= 0)
-			key->pk_algorithm = GNUTLS_PK_RSA;
+		ptr = memmem(data->data, data->size, "PRIVATE KEY-----", sizeof("PRIVATE KEY-----")-1);
 
-		if (result == GNUTLS_E_BASE64_UNEXPECTED_HEADER_ERROR) {
-			/* try for the second header */
-			result =
-			    _gnutls_fbase64_decode(PEM_KEY_DSA, data->data,
-						   data->size, &_data);
+		result = GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE;
 
-			if (result >= 0)
-				key->pk_algorithm = GNUTLS_PK_DSA;
+		if (ptr != NULL) {
+			left = data->size - ((ptrdiff_t)ptr - (ptrdiff_t)data->data);
 
-			if (result ==
-			    GNUTLS_E_BASE64_UNEXPECTED_HEADER_ERROR) {
-				/* try for the second header */
-				result =
-				    _gnutls_fbase64_decode(PEM_KEY_ECC,
-							   data->data,
-							   data->size,
-							   &_data);
-				if (result >= 0)
-					key->pk_algorithm = GNUTLS_PK_EC;
+			if (data->size - left > 15) {
+				ptr -= 15;
+				left += 15;
+			} else {
+				ptr = (char*)data->data;
+				left = data->size;
+			}
+
+			ptr = memmem(ptr, left, "-----BEGIN ", sizeof("-----BEGIN ")-1);
+			if (ptr != NULL) {
+				begin_ptr = (uint8_t*)ptr;
+				left = data->size - ((ptrdiff_t)begin_ptr - (ptrdiff_t)data->data);
+
+				ptr += sizeof("-----BEGIN ")-1;
+			}
+
+			if (ptr != NULL && left > sizeof(PEM_KEY_RSA)) {
+				if (memcmp(ptr, PEM_KEY_RSA, sizeof(PEM_KEY_RSA)-1) == 0) {
+					result =
+					    _gnutls_fbase64_decode(PEM_KEY_RSA, begin_ptr,
+							   	   left, &_data);
+					if (result >= 0)
+						key->pk_algorithm = GNUTLS_PK_RSA;
+				} else if (memcmp(ptr, PEM_KEY_ECC, sizeof(PEM_KEY_ECC)-1) == 0) {
+					result =
+					    _gnutls_fbase64_decode(PEM_KEY_ECC,
+								   begin_ptr,
+								   left,
+								   &_data);
+					if (result >= 0)
+						key->pk_algorithm = GNUTLS_PK_EC;
+				} else if (memcmp(ptr, PEM_KEY_DSA, sizeof(PEM_KEY_DSA)-1) == 0) {
+					result =
+					    _gnutls_fbase64_decode(PEM_KEY_DSA, begin_ptr,
+								   left, &_data);
+					if (result >= 0)
+						key->pk_algorithm = GNUTLS_PK_DSA;
+				}
 			}
 		}
 
@@ -654,9 +676,46 @@ gnutls_x509_privkey_import2(gnutls_x509_privkey_t key,
 {
 	int ret = 0;
 	char pin[GNUTLS_PKCS11_MAX_PIN_LEN];
+	unsigned head_enc = 1;
 
-	if (password == NULL && !(flags & GNUTLS_PKCS_NULL_PASSWORD)) {
+	if (format == GNUTLS_X509_FMT_PEM) {
+		size_t left;
+		char *ptr;
+
+		ptr = memmem(data->data, data->size, "PRIVATE KEY-----", sizeof("PRIVATE KEY-----")-1);
+
+		if (ptr != NULL) {
+			left = data->size - ((ptrdiff_t)ptr - (ptrdiff_t)data->data);
+
+			if (data->size - left > 15) {
+				ptr -= 15;
+				left += 15;
+			} else {
+				ptr = (char*)data->data;
+				left = data->size;
+			}
+
+			ptr = memmem(ptr, left, "-----BEGIN ", sizeof("-----BEGIN ")-1);
+			if (ptr != NULL) {
+				ptr += sizeof("-----BEGIN ")-1;
+				left = data->size - ((ptrdiff_t)ptr - (ptrdiff_t)data->data);
+			}
+
+			if (ptr != NULL && left > sizeof(PEM_KEY_RSA)) {
+				if (memcmp(ptr, PEM_KEY_RSA, sizeof(PEM_KEY_RSA)-1) == 0 ||
+				    memcmp(ptr, PEM_KEY_ECC, sizeof(PEM_KEY_ECC)-1) == 0 ||
+				    memcmp(ptr, PEM_KEY_DSA, sizeof(PEM_KEY_DSA)-1) == 0) {
+				    	head_enc = 0;
+				}
+			}
+		}
+	}
+
+	if (head_enc == 0 || (password == NULL && !(flags & GNUTLS_PKCS_NULL_PASSWORD))) {
 		ret = gnutls_x509_privkey_import(key, data, format);
+		if (ret >= 0)
+			return ret;
+
 		if (ret < 0) {
 			gnutls_assert();
 		}
@@ -1677,8 +1736,6 @@ gnutls_x509_privkey_sign_hash(gnutls_x509_privkey_t key,
  *
  * Returns: On success, %GNUTLS_E_SUCCESS (0) is returned, otherwise a
  *   negative error value.
- *
- * Deprecated: Use gnutls_privkey_sign_data().
  */
 int
 gnutls_x509_privkey_sign_data(gnutls_x509_privkey_t key,
